import os
from ladybug_display_schema.geometry2d import Vector2D, Point2D, Ray2D, \
    LineSegment2D, Polyline2D, Arc2D, Polygon2D, Mesh2D
from ladybug_display_schema.geometry3d import Vector3D, Point3D, Ray3D, \
    Plane, LineSegment3D, Polyline3D, Arc3D, Face3D, Mesh3D, Polyface3D

# target folder where all of the samples live
root = os.path.dirname(os.path.dirname(__file__))
target_folder = os.path.join(root, 'samples', 'geometry')


def test_vector2d():
    file_path = os.path.join(target_folder, 'vector_2d.json')
    Vector2D.parse_file(file_path)


def test_point2d():
    file_path = os.path.join(target_folder, 'point_2d.json')
    Point2D.parse_file(file_path)


def test_ray2d():
    file_path = os.path.join(target_folder, 'ray_2d.json')
    Ray2D.parse_file(file_path)


def test_line_segment_2d():
    file_path = os.path.join(target_folder, 'line_segment_2d.json')
    LineSegment2D.parse_file(file_path)


def test_polyline_2d():
    file_path = os.path.join(target_folder, 'polyline_2d.json')
    Polyline2D.parse_file(file_path)


def test_arc_2d():
    file_path = os.path.join(target_folder, 'arc_2d.json')
    Arc2D.parse_file(file_path)


def test_polygon_2d():
    file_path = os.path.join(target_folder, 'polygon_2d.json')
    Polygon2D.parse_file(file_path)


def test_mesh_2d():
    file_path = os.path.join(target_folder, 'mesh_2d.json')
    Mesh2D.parse_file(file_path)


def test_vector3d():
    file_path = os.path.join(target_folder, 'vector_3d.json')
    Vector3D.parse_file(file_path)


def test_point3d():
    file_path = os.path.join(target_folder, 'point_3d.json')
    Point3D.parse_file(file_path)


def test_ray3d():
    file_path = os.path.join(target_folder, 'ray_3d.json')
    Ray3D.parse_file(file_path)


def test_plane():
    file_path = os.path.join(target_folder, 'plane.json')
    Plane.parse_file(file_path)


def test_line_segment_3d():
    file_path = os.path.join(target_folder, 'line_segment_3d.json')
    LineSegment3D.parse_file(file_path)


def test_polyline_3d():
    file_path = os.path.join(target_folder, 'polyline_3d.json')
    Polyline3D.parse_file(file_path)


def test_arc_3d():
    file_path = os.path.join(target_folder, 'arc_3d.json')
    Arc3D.parse_file(file_path)


def test_face_3d():
    file_path = os.path.join(target_folder, 'face_3d.json')
    Face3D.parse_file(file_path)


def test_mesh_3d():
    file_path = os.path.join(target_folder, 'mesh_3d.json')
    Mesh3D.parse_file(file_path)


def test_polyface_3d():
    file_path = os.path.join(target_folder, 'polyface_3d.json')
    Polyface3D.parse_file(file_path)



"""Schemas for geometry objects in 3D space."""
import math
from typing import List
from pydantic import Field, conlist, constr, conint

from .base import NoExtraBaseModel, Color


class Vector3D(NoExtraBaseModel):
    """A vector object in 3D space."""

    type: constr(regex='^Vector3D$') = 'Vector3D'

    x: float = Field(
        ...,
        description='Number for X coordinate.'
    )

    y: float = Field(
        ...,
        description='Number for Y coordinate.'
    )

    z: float = Field(
        ...,
        description='Number for Z coordinate.'
    )


class Point3D(NoExtraBaseModel):
    """A point object in 3D space."""

    type: constr(regex='^Point3D$') = 'Point3D'

    x: float = Field(
        ...,
        description='Number for X coordinate.'
    )

    y: float = Field(
        ...,
        description='Number for Y coordinate.'
    )

    z: float = Field(
        ...,
        description='Number for Z coordinate.'
    )


class Ray3D(NoExtraBaseModel):
    """A ray object in 3D space."""

    type: constr(regex='^Ray3D$') = 'Ray3D'

    p: List[float] = Field(
        ...,
        description="Ray base point as 3 (x, y, z) values.",
        min_items=3,
        max_items=3
    )

    v: List[float] = Field(
        ...,
        description="Ray direction vector as 3 (x, y, z) values.",
        min_items=3,
        max_items=3
    )


class Plane(NoExtraBaseModel):
    """A plane object."""

    type: constr(regex='^Plane$') = 'Plane'

    n: List[float] = Field(
        ...,
        description="Plane normal as 3 (x, y, z) values.",
        min_items=3,
        max_items=3
    )

    o: List[float] = Field(
        ...,
        description="Plane origin as 3 (x, y, z) values",
        min_items=3,
        max_items=3
    )

    x: List[float] = Field(
        default=None,
        description="Plane x-axis as 3 (x, y, z) values. If None, it is autocalculated.",
        min_items=3,
        max_items=3
    )


class LineSegment3D(NoExtraBaseModel):
    """A single line segment face in 3D space."""

    type: constr(regex='^LineSegment3D$') = 'LineSegment3D'

    p: List[float] = Field(
        ...,
        description="Line segment base point as 3 (x, y, z) values.",
        min_items=3,
        max_items=3
    )

    v: List[float] = Field(
        ...,
        description="Line segment direction vector as 3 (x, y, z) values.",
        min_items=3,
        max_items=3
    )


class Polyline3D(NoExtraBaseModel):
    """A polyline in 3D space."""

    type: constr(regex='^Polyline3D$') = 'Polyline3D'

    vertices: List[conlist(float, min_items=3, max_items=3)] = Field(
        ...,
        min_items=3,
        description='A list of points representing the the vertices of the polyline. '
        'The list should include at least 3 points and each point '
        'should be a list of 3 (x, y, z) values.'
    )

    interpolated: bool = Field(
        False,
        description='A boolean to note whether the polyline should be interpolated '
        'between the input vertices when it is translated to other interfaces.'
    )


class Arc3D(NoExtraBaseModel):
    """A single arc or circle in 3D space."""

    type: constr(regex='^Arc3D$') = 'Arc3D'

    plane: Plane = Field(
        ...,
        description='A Plane in which the arc lies with an origin representing the '
        'center of the circle for the arc.'
    )

    radius: float = Field(
        ...,
        description='A number representing the radius of the arc.'
    )

    a1: float = Field(
        default=0,
        ge=0,
        le=math.pi * 2,
        description='A number between 0 and 2 * pi for the start angle of the arc.'
    )

    a2: float = Field(
        default=math.pi * 2,
        ge=0,
        le=math.pi * 2,
        description='A number between 0 and 2 * pi for the end angle of the arc.'
    )


class Face3D(NoExtraBaseModel):
    """A single planar face in 3D space."""

    type: constr(regex='^Face3D$') = 'Face3D'

    boundary: List[conlist(float, min_items=3, max_items=3)] = Field(
        ...,
        min_items=3,
        description='A list of points representing the outer boundary vertices of '
        'the face. The list should include at least 3 points and each point '
        'should be a list of 3 (x, y, z) values.'
    )

    holes: List[conlist(conlist(float, min_items=3, max_items=3), min_items=3)] = Field(
        default=None,
        description='Optional list of lists with one list for each hole in the face.'
        'Each hole should be a list of at least 3 points and each point a list '
        'of 3 (x, y, z) values. If None, it will be assumed that there are no '
        'holes in the face.'
    )

    plane: Plane = Field(
        default=None,
        description='Optional Plane indicating the plane in which the face exists.'
        'If None, the plane will usually be derived from the boundary points.'
    )


class Mesh3D(NoExtraBaseModel):
    """A mesh in 3D space."""

    type: constr(regex='^Mesh3D$') = 'Mesh3D'

    vertices: List[conlist(float, min_items=3, max_items=3)] = Field(
        ...,
        min_items=3,
        description='A list of points representing the vertices of the mesh. '
        'The list should include at least 3 points and each point '
        'should be a list of 3 (x, y, z) values.'
    )

    faces: List[conlist(conint(ge=0), min_items=3, max_items=4)] = Field(
        ...,
        min_items=1,
        description='A list of lists with each sub-list having either 3 or 4 '
        'integers. These integers correspond to indices within the list of vertices.'
    )

    colors: List[Color] = Field(
        None,
        description='An optional list of colors that correspond to either the faces '
        'of the mesh or the vertices of the mesh.'
    )


class PolyfaceEdgeInfo(NoExtraBaseModel):
    """Optional edge information for Polyface."""

    edge_indices: List[conlist(conint(ge=0), min_items=2, max_items=2)] = Field(
        ...,
        min_items=3,
        description='An array objects that each contain two integers. These integers '
        'correspond to indices within the vertices list and each tuple represents '
        'a line segment for an edge of the polyface.'
    )

    edge_types: List[conint(ge=0)] = Field(
        ...,
        min_items=3,
        description='An array of integers for each edge that parallels the edge_indices '
        'list. An integer of 0 denotes a naked edge, an integer of 1 denotes an '
        'internal edge. Anything higher is a non-manifold edge.'
    )


class Polyface3D(NoExtraBaseModel):
    """A Polyface in 3D space."""

    type: constr(regex='^Polyface3D$') = 'Polyface3D'

    vertices: List[conlist(float, min_items=3, max_items=3)] = Field(
        ...,
        min_items=3,
        description='A list of points representing the vertices of the polyface. '
        'The list should include at least 3 points and each point '
        'should be a list of 3 (x, y, z) values.'
    )

    face_indices: List[
        conlist(conlist(conint(ge=0), min_items=3), min_items=1)
    ] = Field(
        ...,
        min_items=1,
        description='A list of lists with one list for each face of the polyface. '
        'Each face list must contain at least one sub-list of integers corresponding '
        'to indices within the vertices list. Additional sub-lists of integers may '
        'follow this one such that the first sub-list denotes the boundary of the '
        'face while each subsequent sub-list denotes a hole in the face.'
    )

    edge_information: PolyfaceEdgeInfo = Field(
        default=None,
        description='Optional edge information, which will speed up the '
        'creation of the Polyface object if it is available but should be left '
        'as None if it is unknown. If None, edge_information will be computed '
        'from the vertices and face_indices inputs.'
    )


class Sphere(NoExtraBaseModel):
    """A sphere object."""

    type: constr(regex='^Sphere$') = 'Sphere'

    center: List[float] = Field(
        ...,
        description='The center of the sphere as 3 (x, y, z) values.',
        min_items=3,
        max_items=3
    )

    radius: float = Field(
        ...,
        gt=0,
        description='A number representing the radius of the sphere.'
    )


class Cone(NoExtraBaseModel):
    """A cone object."""

    type: constr(regex='^Cone$') = 'Sphere'

    vertex: List[float] = Field(
        ...,
        description='The point at the tip of the cone as 3 (x, y, z) values.',
        min_items=3,
        max_items=3
    )

    axis: List[float] = Field(
        ...,
        description='The vector representing the direction of the cone as 3 (x, y, z) '
        'values. The vector extends from the vertex to the center of the base.',
        min_items=3,
        max_items=3
    )

    angle: float = Field(
        ...,
        gt=0,
        lt=math.pi / 2,
        description='An angle in radians representing the half angle between '
        'the axis and the surface.'
    )


class Cylinder(NoExtraBaseModel):
    """A cylinder object."""

    type: constr(regex='^Cylinder$') = 'Cylinder'

    center: List[float] = Field(
        ...,
        description='The center of the bottom base of the cylinder as '
        '3 (x, y, z) values.',
        min_items=3,
        max_items=3
    )

    axis: List[float] = Field(
        ...,
        description='The vector representing the direction of the cylinder as '
        '3 (x, y, z) values. The vector extends from the bottom base center to '
        'the top base center.',
        min_items=3,
        max_items=3
    )

    radius: float = Field(
        ...,
        gt=0,
        description='A number representing the radius of the cylinder.'
    )



"""Schemas for geometric display objects in 3D space."""
from typing import List
from pydantic import Field, constr, root_validator

from .base import DisplayBaseModel, SingleColorBase, LineCurveBase, \
    Color, DisplayModes
from .geometry3d import Vector3D, Point3D, Ray3D, Plane, LineSegment3D, \
    Polyline3D, Arc3D, Face3D, Mesh3D, Polyface3D, Sphere, Cone, Cylinder


class DisplayVector3D(SingleColorBase):
    """A point object in 3D space with display properties."""

    type: constr(regex='^DisplayVector3D$') = 'DisplayVector3D'

    geometry: Vector3D = Field(
        ...,
        description='Vector3D for the geometry.'
    )


class DisplayPoint3D(SingleColorBase):
    """A point object in 3D space with display properties."""

    type: constr(regex='^DisplayPoint3D$') = 'DisplayPoint3D'

    geometry: Point3D = Field(
        ...,
        description='Point3D for the geometry.'
    )


class DisplayRay3D(SingleColorBase):
    """A ray object in 3D space with display properties."""

    type: constr(regex='^DisplayRay3D$') = 'DisplayRay3D'

    geometry: Ray3D = Field(
        ...,
        description='Ray3D for the geometry.'
    )


class DisplayPlane(SingleColorBase):
    """A plane object with display properties."""

    type: constr(regex='^DisplayPlane$') = 'DisplayPlane'

    geometry: Plane = Field(
        ...,
        description='Plane for the geometry.'
    )

    show_axes: bool = Field(
        False,
        description='A boolean to note whether the plane should be displayed '
        'with XY axes instead of just an origin point and a normal vector.'
    )

    show_grid: bool = Field(
        False,
        description='A boolean to note whether the plane should be displayed '
        'with a grid.'
    )


class DisplayLineSegment3D(LineCurveBase):
    """A single line segment face in 3D space with display properties."""

    type: constr(regex='^DisplayLineSegment3D$') = 'DisplayLineSegment3D'

    geometry: LineSegment3D = Field(
        ...,
        description='LineSegment3D for the geometry.'
    )


class DisplayPolyline3D(LineCurveBase):
    """A polyline in 3D space with display properties."""

    type: constr(regex='^DisplayPolyline3D$') = 'DisplayPolyline3D'

    geometry: Polyline3D = Field(
        ...,
        description='Polyline3D for the geometry.'
    )


class DisplayArc3D(LineCurveBase):
    """A single arc or circle in 3D space with display properties."""

    type: constr(regex='^DisplayArc3D$') = 'DisplayArc3D'

    geometry: Arc3D = Field(
        ...,
        description='Arc3D for the geometry.'
    )


class DisplayFace3D(SingleColorBase):
    """A single planar face in 3D space with display properties."""

    type: constr(regex='^DisplayFace3D$') = 'DisplayFace3D'

    geometry: Face3D = Field(
        ...,
        description='Face3D for the geometry.'
    )

    display_mode: DisplayModes = Field(
        DisplayModes.shaded,
        description='Text to indicate the display mode (shaded, wireframe, '
        'etc.). The DisplayModes enumeration contains all acceptable types.'
    )


class DisplayMesh3D(DisplayBaseModel):
    """A mesh in 3D space with display properties."""

    type: constr(regex='^DisplayMesh3D$') = 'DisplayMesh3D'

    geometry: Mesh3D = Field(
        ...,
        description='Mesh3D for the geometry.'
    )

    colors: List[Color] = Field(
        ...,
        description='A list of colors that correspond to either the faces '
        'of the mesh or the vertices of the mesh. It can also be a single color '
        'for the entire mesh.'
    )

    display_mode: DisplayModes = Field(
        DisplayModes.shaded,
        description='Text to indicate the display mode (shaded, wireframe, '
        'etc.). The DisplayModes enumeration contains all acceptable types.'
    )

    @root_validator
    def check_colors_align(cls, values):
        """Check that there are an acceptable number of colors for the mesh."""
        colors, geo = values.get('colors'), values.get('geometry')
        if colors is None:
            return values
        if len(colors) in (1, len(geo.faces), len(geo.vertices)):
            return values
        else:
            raise ValueError(
                'Number of colors ({}) does not match the number of mesh faces '
                '({}) nor the number of vertices ({}).'.format(
                    len(colors), len(geo.faces), len(geo.vertices))
            )


class DisplayPolyface3D(DisplayBaseModel):
    """A Polyface in 3D space with display properties."""

    type: constr(regex='^DisplayPolyface3D$') = 'DisplayPolyface3D'

    geometry: Polyface3D = Field(
        ...,
        description='Polyface3D for the geometry.'
    )

    colors: List[Color] = Field(
        ...,
        description='A list of colors that correspond to either the faces of the '
        'polyface or the vertices of the polyface. It can also be a single color '
        'for the entire polyface.'
    )

    display_mode: DisplayModes = Field(
        DisplayModes.shaded,
        description='Text to indicate the display mode (shaded, wireframe, '
        'etc.). The DisplayModes enumeration contains all acceptable types.'
    )

    @root_validator
    def check_colors_align(cls, values):
        """Check that there are an acceptable number of colors for the mesh."""
        colors, geo = values.get('colors'), values.get('geometry')
        if colors is None:
            return values
        if len(colors) in (1, len(geo.face_indices), len(geo.vertices)):
            return values
        else:
            raise ValueError(
                'Number of colors ({}) does not match the number of mesh faces '
                '({}) nor the number of vertices ({}).'.format(
                    len(colors), len(geo.face_indices), len(geo.vertices))
            )


class DisplaySphere(SingleColorBase):
    """A sphere object with display properties."""

    type: constr(regex='^DisplaySphere$') = 'DisplaySphere'

    geometry: Sphere = Field(
        ...,
        description='Sphere for the geometry.'
    )

    display_mode: DisplayModes = Field(
        DisplayModes.shaded,
        description='Text to indicate the display mode (shaded, wireframe, '
        'etc.). The DisplayModes enumeration contains all acceptable types.'
    )


class DisplayCone(SingleColorBase):
    """A cone object with display properties."""

    type: constr(regex='^DisplayCone$') = 'DisplayCone'

    geometry: Cone = Field(
        ...,
        description='Cone for the geometry.'
    )

    display_mode: DisplayModes = Field(
        DisplayModes.shaded,
        description='Text to indicate the display mode (shaded, wireframe, '
        'etc.). The DisplayModes enumeration contains all acceptable types.'
    )


class DisplayCylinder(SingleColorBase):
    """A cylinder object with display properties."""

    type: constr(regex='^DisplayCylinder$') = 'DisplayCylinder'

    geometry: Cylinder = Field(
        ...,
        description='Cylinder for the geometry.'
    )

    display_mode: DisplayModes = Field(
        DisplayModes.shaded,
        description='Text to indicate the display mode (shaded, wireframe, '
        'etc.). The DisplayModes enumeration contains all acceptable types.'
    )



"""Schemas for geometric display objects in 2D space."""
from typing import List
from pydantic import Field, constr, root_validator

from .base import DisplayBaseModel, SingleColorBase, LineCurveBase, \
    Color, DisplayModes
from .geometry2d import Vector2D, Point2D, Ray2D, LineSegment2D, \
    Polyline2D, Arc2D, Polygon2D, Mesh2D


class DisplayVector2D(SingleColorBase):
    """A point object in 2D space with display properties."""

    type: constr(regex='^DisplayVector2D$') = 'DisplayVector2D'

    geometry: Vector2D = Field(
        ...,
        description='Vector2D for the geometry.'
    )


class DisplayPoint2D(SingleColorBase):
    """A point object in 2D space with display properties."""

    type: constr(regex='^DisplayPoint2D$') = 'DisplayPoint2D'

    geometry: Point2D = Field(
        ...,
        description='Point2D for the geometry.'
    )


class DisplayRay2D(SingleColorBase):
    """A ray object in 2D space with display properties."""

    type: constr(regex='^DisplayRay2D$') = 'DisplayRay2D'

    geometry: Ray2D = Field(
        ...,
        description='Ray2D for the geometry.'
    )


class DisplayLineSegment2D(LineCurveBase):
    """A single line segment face in 2D space with display properties."""

    type: constr(regex='^DisplayLineSegment2D$') = 'DisplayLineSegment2D'

    geometry: LineSegment2D = Field(
        ...,
        description='LineSegment2D for the geometry.'
    )


class DisplayPolyline2D(LineCurveBase):
    """A polyline in 2D space with display properties."""

    type: constr(regex='^DisplayPolyline2D$') = 'DisplayPolyline2D'

    geometry: Polyline2D = Field(
        ...,
        description='Polyline2D for the geometry.'
    )


class DisplayArc2D(LineCurveBase):
    """A single arc or circle in 2D space with display properties."""

    type: constr(regex='^DisplayArc2D$') = 'DisplayArc2D'

    geometry: Arc2D = Field(
        ...,
        description='Arc2D for the geometry.'
    )


class DisplayPolygon2D(SingleColorBase):
    """A single polygon in 2D space with display properties."""

    type: constr(regex='^DisplayPolygon2D$') = 'DisplayPolygon2D'

    geometry: Polygon2D = Field(
        ...,
        description='Polygon2D for the geometry.'
    )

    display_mode: DisplayModes = Field(
        DisplayModes.shaded,
        description='Text to indicate the display mode (shaded, wireframe, '
        'etc.). The DisplayModes enumeration contains all acceptable types.'
    )


class DisplayMesh2D(DisplayBaseModel):
    """A mesh in 2D space with display properties."""

    type: constr(regex='^DisplayMesh2D$') = 'DisplayMesh2D'

    geometry: Mesh2D = Field(
        ...,
        description='Mesh2D for the geometry.'
    )

    colors: List[Color] = Field(
        ...,
        description='A list of colors that correspond to either the faces '
        'of the mesh or the vertices of the mesh. It can also be a single color '
        'for the entire mesh.'
    )

    display_mode: DisplayModes = Field(
        DisplayModes.shaded,
        description='Text to indicate the display mode (shaded, wireframe, '
        'etc.). The DisplayModes enumeration contains all acceptable types.'
    )

    @root_validator
    def check_colors_align(cls, values):
        """Check that there are an acceptable number of colors for the mesh."""
        colors, geo = values.get('colors'), values.get('geometry')
        if colors is None:
            return values
        if len(colors) in (1, len(geo.faces), len(geo.vertices)):
            return values
        else:
            raise ValueError(
                'Number of colors ({}) does not match the number of mesh faces '
                '({}) nor the number of vertices ({}).'.format(
                    len(colors), len(geo.faces), len(geo.vertices))
            )
